import os.path
from zozol import decode_ber, base as asn1, markers as m
from zozol.schemas.pkcs7_dstszi import ContentInfo
from zozol.util import u
from gost89 import gosthash

def here(fname):
    dirname, _ = os.path.split(__file__)
    return os.path.join(dirname, fname)

def test_pkcs7():
    CONTENT = """IGNORE THIS FILE.
This file does nothing, contains no useful data, and might go away in
future releases.  Do not depend on this file or its contents.
"""

    data = open(here('signed1.r'), 'rb').read()
    msg = ContentInfo.stream(bytearray(data), len(data), decode_ber)
    assert str(msg.content.contentInfo.content) == CONTENT
    assert msg.content.signerInfos[0].sid.serialNumber.value == 359272175317388400160838857906663248925214184704
    sign = bytearray(b'\x90\x96Sum/\x0f\x97\x02I\x8e\t%\xe2\xd1\xa5\x0e\x8a\xa6\x17\xd3\x16\x80\xc4E\x9c\xb4A\xef\x8ckb\x0e\x86\xde\x8e7\x03k\xa5\x99\xc6\x17\x19r\x9a;u3\x92\x16\xa2P\x83\x9bO\xf8g\xe4\xe1;yc\r')
    assert msg.content.signerInfos[0].encryptedDigest.value == u(sign, 'latin1')


def test_pkcs7_certtax():
    data = open(here('signed1.r'), 'rb').read()
    msg = ContentInfo.stream(bytearray(data), len(data), decode_ber)
    x509 = msg.content.certificates[0]
    assert x509.subject_edrpou == '3225813873'
    assert x509.subject_drfo == '3225813873'


def test_implicit():
    class X(asn1.Seq):
        fields = [
            ('b', m.Explicit(tag=0, typ=asn1.OctStr))
        ]

    data = bytearray(b'\x30\x07\xA0\x05\x04\x03\x31\x32\x33')
    x = X.stream(data, len(data), decode_ber)
    assert str(x.b) == '123'


def test_optional():
    class X(asn1.Seq):
        fields = [
            ('a', m.Optional(asn1.Int)),
            ('b', m.Explicit(tag=0, typ=asn1.OctStr))
        ]

    data = bytearray(b'\x30\x07\xA0\x05\x04\x03\x31\x32\x33')
    x = X.stream(data, len(data), decode_ber)
    assert str(x.b) == '123'
    assert not hasattr(x, 'a')


def test_optional_last():
    class A(asn1.Seq):
        fields = [
            ('key', asn1.Bool),
            ('opt', m.Optional(asn1.Bool)),
        ]

    data = bytearray(b'\x30\x03\x01\x01\xff')
    a = A.stream(data, decode_fn=decode_ber)
    assert a.key.value is True
    assert not hasattr(a, 'opt')


def test_default():
    class A(asn1.Seq):
        fields = [
            ('key', asn1.Bool),
            ('opt', m.Default(value='default', typ=asn1.OctStr)),
        ]

    data = bytearray(b'\x30\x03\x01\x01\xff')
    a = A.stream(data, decode_fn=decode_ber)
    assert a.key.value is True
    assert a.opt.value == 'default'


def test_choice():
    class IS(m.Choice):
        types = [
            asn1.Int,
            asn1.OctStr,
        ]

    class A(asn1.Seq):
        fields = [
            ('value', IS),
        ]

    data = bytearray(b'\x30\x03\x02\x01\xff')
    a = A.stream(data, decode_fn=decode_ber)
    assert isinstance(a.value, asn1.Int)
    assert a.value.value == 255

    data = bytearray(b'\x30\x03\x04\x01\x0d')
    a = A.stream(data, decode_fn=decode_ber)
    assert isinstance(a.value, asn1.OctStr)
    assert a.value.value == '\x0d'


def test_seq_int():
    class B(asn1.Int):
        pass

    data = bytearray(b'\x02\x02\xFF\x01')
    b = B.stream(data, decode_fn=decode_ber)
    assert b.value == 0xFF01, str(b)


def test_explicit():
    class E(asn1.Seq):
        fields = [
            ('name', m.Explicit(tag=2, typ=asn1.OctStr))
        ]

    data = bytearray(b'\x30\x05\xa2\x03\x04\x01\x58')
    e = E.stream(data, decode_fn=decode_ber)
    assert str(e.name) == 'X'


def test_explicit_raw():
    data = bytearray(b'\xa2\x03\x04\x01\x58')
    source = decode_ber(data, len(data))
    tag, cls, content = next(source)
    assert tag == 2

    source = decode_ber(content.data, content.tlen, content.off)
    tag, cls, content = next(source)
    assert tag == 4
    assert cls == 0, cls
    assert str(content) == 'X'


def test_bool():
    data = bytearray(b'\x01\x01\xff')
    b = asn1.Bool.stream(data, decode_fn=decode_ber)
    assert b.value is True


def test_int():
    data = bytearray(b'\x02\x02\x00\x80')
    d = asn1.Int.stream(data, decode_fn=decode_ber)
    assert d.value == 128


def test_hashsum():
    data = open(here('signed1.r'), 'rb').read()
    msg = ContentInfo.stream(bytearray(data), len(data), decode_ber)
    attrs = msg.content.signerInfos[0].authenticatedAttributes
    dgst = attrs['1.2.840.113549.1.9.4'][0].data.value.encode('latin1')
    cdgst = gosthash(msg.content.contentInfo.content.value.encode('latin1'))

    assert cdgst == dgst

    dgst = attrs.messageDigest.value.encode('latin1')
    assert cdgst == dgst



from __future__ import print_function
import sys
from collections import namedtuple
import base64
try:
    import StringIO as io
except ImportError:
    import io
import struct

Document = namedtuple('Document', ('label', 'body', 'headers'))

if sys.version_info.major == 2:
    def u(data, cp='latin1'):
        return unicode(bytes(data), cp)
else:
    u = str


class LabelError(ValueError):
    pass


def to_pem(data, name):
    ret = io.StringIO()
    ret.write('-----BEGIN {}-----\n'.format(name.upper()))
    for x in range(0, len(data), 48):
        ret.write(u(base64.b64encode(data[x:x+48]), 'latin1'))
        ret.write('\n')
    ret.write('-----END {}-----\n'.format(name.upper()))
    return ret.getvalue()


def strip_header(data, off):
    section = off
    while data[off] and (off - section) < 20:
        off += 1

    off += 1
    label = u(data[section:off], 'latin1')
    (tlen,) = struct.unpack_from('<I', data, off)
    off += 4
    return label, off, tlen


def decode_hlist(hl, cp):
    return {
        u(k, 'latin1'): u(bytes(v), cp)
        for k, v in hl
    }


def decode_headers(data):
    ret = []
    is_win = False
    for line in data.split(b'\r\n'):
        if b'=' not in line:
            continue

        k, v = line.split(b'=', 1)
        ret.append((k, v))
        if k == b'ENCODING' and v == b'WIN':
            is_win = True

    try:
        return decode_hlist(ret, 'cp1251' if is_win else 'utf-8')
    except UnicodeDecodeError:
        return decode_hlist(ret, 'cp1251')


def decode_transport(data):
    if not isinstance(data, bytearray):
        data = bytearray(data)

    off = 0
    headers = {}
    label, off, tlen = strip_header(data, off)

    if label == 'TRANSPORTABLE\0':
        headers = decode_headers(data[off:off+tlen])
        label, off, tlen = strip_header(data, off + tlen)
    if label == 'CERTCRYPT\0':
        headers["CERT_PEM"] = to_pem(data[off:off+tlen], 'CERTIFICATE')
        label, off, tlen = strip_header(data, off + tlen)

    if label[3:] in ('_SIGN\0', '_CRYPT\0'):
        pass
    else:
        raise LabelError("Unknown label {}".format(repr(label)))

    return Document(label=label[:-1], body=data[off:off+tlen], headers=headers)



import types
from . import base
from . import markers
from . rewindable import Rewindable

tags = {
    0x01: base.Bool,
    0x06: base.ObjId,
    0x03: base.BitStr,
    0x04: base.OctStr,
    0x0C: base.Utf8Str,
    0x13: base.PrintStr,
    0x02: base.Int,
    0x11: base.SetOf,
    0x17: base.Time,
}


def decode_len(data, off):
    tlen = data[off]
    off += 1
    hlen = 1
    if tlen & 0x80:
        lenlen = tlen & 0x7F
        hlen += lenlen
        tlen = 0
        while lenlen:
            tlen = tlen << 8
            tlen |= data[off]
            off += 1
            lenlen -= 1

    return tlen, hlen


def decode(data, avail, off=0):
    while avail > 0:
        hlen = 1
        tag = data[off] & 0x1F
        cls = data[off] >> 6
        off += 1

        tlen, llen = decode_len(data, off)
        off += llen
        hlen += llen

        ret = None
        if cls == 0:  # u
            if tag == 0x10 or tag == 0x11:
                ret = Rewindable(fn=decode, data=data, tlen=tlen, off=off)
            else:
                ret = tags[tag](data[off:off+tlen])
        elif cls == 2:
            ret = markers.Tagged(tag, data, off, tlen)

        if ret is None:
            ret = data[off:off+tlen]

        avail -= tlen + hlen
        off += tlen

        yield tag, cls, ret


def encode_tag(tag, cls, content, container, ident=0):
    container.append(tag | cls << 6)
    tlen = len(content)
    if tlen < 128:
        container.append(tlen)
    else:
        len_bytes = []
        while tlen:
            len_bytes.append(tlen & 0xFF)
            tlen >>= 8

        container.append(0x80 | len(len_bytes))
        while len_bytes:
            container.append(len_bytes.pop())

    for byte in content:
        container.append(byte)

    return container


def encode(inp, ident=0):
    ret = bytearray()
    for (tag, cls, content) in inp:
        if type(content) in [types.GeneratorType, list, tuple]:
            content = encode(content, ident=ident+1)
        if tag == 0x10 or tag == 0x11:
            tag = tag | 0x20

        encode_tag(tag, cls, content, ret, ident=ident+1)

    return ret



import os.path
from zozol import decode_ber, encode_ber, encode_ber_tag, base as asn1, markers as m
from zozol.schemas.pkcs7_dstszi import ContentInfo

def here(fname):
    dirname, _ = os.path.split(__file__)
    return os.path.join(dirname, fname)

def test_encode_ocsttr():
    data = encode_ber_tag(0x04, 0, bytearray('123', 'latin1'), bytearray())
    assert data == b'\x04\x03\x31\x32\x33'


def test_encode_ocsttr_long():
    expected = bytearray(b'\x04\x81\x8033333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333')
    data = encode_ber_tag(0x04, 0, bytearray('3' * 128, 'latin1'), bytearray())
    assert data == expected, (repr(data), repr(expected))


def test_encode_ocsttr_long2():
    expected = bytearray(b'\x04\x82\x01\x0133333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333')
    data = encode_ber_tag(0x04, 0, bytearray('3' * 257, 'latin1'), bytearray())
    assert data == expected, (repr(data), repr(expected))


def test_encode_seq():
    seq = (
        (0x30, 0, (
            (0x04, 0, bytearray('123', 'latin1')),
            (0, 2, bytearray('999', 'latin1')),
        ))
    )
    data = encode_ber((seq,))
    assert len(data) == 2 + 2 + 2 + 3 + 3
    assert data[0] == 0x30
    assert data[1] == 2 + 2 + 3 + 3
    assert data[2] == 0x04
    assert data[3] == 3
    assert data[4:7] == bytearray('123', 'latin1')
    assert data[7] == 0x80
    assert data[8] == 3
    assert data[9:12] == bytearray('999', 'latin1')


def test_encode_schema():
    class X(asn1.Seq):
        fields = [
            ('b', asn1.OctStr),
            ('c', asn1.Int),
        ]

    x = X()
    x.b = asn1.OctStr(value='123')
    x.c = asn1.Int(value=2227)

    data = x.to_stream(encode_fn=encode_ber)
    assert bytes(data) == b'\x30\x09\x04\x03\x31\x32\x33\x02\x02\x08\xb3'


def test_encode_implicit():
    class X(asn1.Seq):
        fields = [
            ('b', m.Implicit(tag=0, typ=asn1.OctStr)),
            ('c', asn1.Int),
        ]

    x = X()
    x.b = asn1.OctStr(value='123')
    x.c = asn1.Int(value=2227)

    data = x.to_stream(encode_fn=encode_ber)
    assert bytes(data) == b'\x30\x09\x80\x03\x31\x32\x33\x02\x02\x08\xb3'


def test_encode_explicit():
    class X(asn1.Seq):
        fields = [
            ('b', m.Explicit(tag=0, typ=asn1.OctStr)),
            ('c', asn1.Int),
        ]

    x = X()
    x.b = asn1.OctStr(value='123')
    x.c = asn1.Int(value=2227)

    data = x.to_stream(encode_fn=encode_ber)
    assert bytes(data) == b'\x30\x0B\xA0\x05\x04\x03\x31\x32\x33\x02\x02\x08\xB3'


def test_encode_default():
    class X(asn1.Seq):
        fields = [
            ('c', m.Default(value=2, typ=asn1.Int)),
        ]

    x = X()

    data = x.to_stream(encode_fn=encode_ber)
    assert bytes(data) == b'\x30\x00'

    x.c = asn1.Int(value=2)

    data = x.to_stream(encode_fn=encode_ber)
    assert bytes(data) == b'\x30\x00'

    x.c = asn1.Int(value=3)

    data = x.to_stream(encode_fn=encode_ber)
    assert bytes(data) == b'\x30\x03\x02\x01\x03'


def test_encode_attrs():
    data = open(here('signed1.r'), 'rb').read()
    cattr_data = bytearray(open(here('signed1.attrs'), 'rb').read())

    msg = ContentInfo.stream(bytearray(data), len(data), decode_ber)
    attrs = msg.content.signerInfos[0].authenticatedAttributes
    attr_data = attrs.to_stream(encode_fn=encode_ber)
    cattr_data[0] = 0x31
    assert attr_data == cattr_data


def test_encode_x509():
    data = open(here('signed1.r'), 'rb').read()
    cx509_data = bytearray(open(here('signed1.x509'), 'rb').read())

    msg = ContentInfo.stream(bytearray(data), decode_fn=decode_ber)
    x509 = msg.content.certificates[0]
    x509_data = x509.to_stream(encode_fn=encode_ber)
    assert x509_data == cx509_data


def test_detach():
    data = open(here('signed1.r'), 'rb').read()
    msg = ContentInfo.stream(bytearray(data), decode_fn=decode_ber)
    recode = msg.to_stream(encode_fn=encode_ber)
    assert data == bytes(recode)

    del msg.content.contentInfo.content
    detached_sign_data = msg.to_stream(encode_fn=encode_ber)
    assert detached_sign_data[55] == 0xA0, repr(detached_sign_data[55:60])
    msg = ContentInfo.stream(detached_sign_data, decode_fn=decode_ber)
    assert not hasattr(msg.content.contentInfo, 'content')


def test_choice():
    class IS(m.Choice):
        types = [
            asn1.Int,
            asn1.OctStr,
        ]

    class A(asn1.Seq):
        fields = [
            ('value', IS),
        ]

    a = A()
    a.value = asn1.Int(value=255)
    data = a.to_stream(encode_fn=encode_ber)
    expect_data = bytearray(b'\x30\x03\x02\x01\xff')
    assert data == expect_data

    a = A.stream(data, decode_fn=decode_ber)
    a.value = asn1.OctStr(value='\x0d')
    data = a.to_stream(encode_fn=encode_ber)
    expect_data = bytearray(b'\x30\x03\x04\x01\x0d')
    assert data == expect_data



from . ber import decode as decode_ber, \
                  encode as encode_ber, \
                  encode_tag as encode_ber_tag
from . util import to_pem, decode_transport

__all__ = [
    'decode_ber',
    'encode_ber',
    'encode_ber_tag',
    'to_pem',
    'decode_transport',
]



import os.path
from zozol import to_pem


def here(fname):
    dirname, _ = os.path.split(__file__)
    return os.path.join(dirname, fname)


def test_pem():
    x509_data = open(here('signed1.x509'), 'rb').read()
    p509_data = open(here('signed1.x509.pem'), 'r').read()
    pem_data = to_pem(x509_data, 'certificate')
    assert pem_data == p509_data



from .. import base as asn1
from .. markers import Implicit, Explicit, Optional, Default


class AttributeTypeAndValue(asn1.Seq):
    fields = [
        ('type', asn1.ObjId),
        ('value', asn1.Any),
    ]

class RelativeDistinguishedName(asn1.SetOf):
    typ = AttributeTypeAndValue

class RDNSequence(asn1.SeqOf):
    typ = RelativeDistinguishedName

class Name(RDNSequence):
    pass


class DigestAlgorithmIdentifier(asn1.Seq):
    fields = [
        ('algorithm', asn1.ObjId),
        ('parameters', Optional(asn1.Any)),
    ]


class UniqueIdentifier(asn1.BitStr):
    pass


class SubjectPublicKeyInfo(asn1.Seq):
    fields = [
        ('algorithm', DigestAlgorithmIdentifier),
        ('subjectPublicKey', asn1.BitStr),
        
    ]


class Validity(asn1.Seq):
    fields = [
        ('notBefore', asn1.Time),
        ('notAfter', asn1.Time),
    ]


class Strings(asn1.SetOf):
    typ = asn1.PrintStr


class TaxId(asn1.Seq):
    fields = [
        ('attribute', asn1.ObjId),
        ('value', Strings),
    ]

class TaxIds(asn1.SeqOf):
    typ = TaxId

    @property
    def value(self):
        ret = []
        for el in self.elements:
            ret.append((str(el.attribute), el.value[0]))
        return dict(ret)


class ExtnValue(asn1.OctStr):
    TYPES = {
        "2.5.29.14": asn1.OctStr,
        "2.5.29.15": asn1.BitStr,
        "2.5.29.9": TaxIds,
    }

    @classmethod
    def from_data(cls, data, decode_fn, parent, **kwargs):
        desc = cls.TYPES.get(str(parent.extnID))
        if desc is None:
            return asn1.OctStr.from_data(data, decode_fn)

        data = bytearray(data.value, 'latin1')
        return desc.stream(data, decode_fn=decode_fn)

    @classmethod
    def to_data(cls, data, parent):
        desc = cls.TYPES.get(str(parent.extnID))
        if desc is None:
            return data

        return (cls.tag, 0, (data,))


class Extension(asn1.Seq):
    fields = [
        ('extnID', asn1.ObjId),
        ('critical', Default(value=False, typ=asn1.Bool)),
        ('extnValue', ExtnValue),
    ]


class Extensions(asn1.SeqOf):
    typ = Extension

    @property
    def ext(self):
        try:
            return self._ext
        except AttributeError:
            self._ext = self.ext_cache()

        return self._ext

    def ext_cache(self):
        ret = {}
        for ext in self.elements:
            ret[str(ext.extnID)] = ext.extnValue.value

        return ret

    def __getitem__(self, idx):
        return self.ext[idx]

    def __repr__(self):
        return '<Extensions {}>'.format(self.ext.keys())

class TBSCertificate(asn1.Seq):
    fields = [
        ('version', Explicit(tag=0, typ=asn1.Int)),
        ('serialNumber', asn1.Int),
        ('signature', DigestAlgorithmIdentifier),
        ('issuer', Name),
        ('validity', Validity),
        ('subject', Name),
        ('subjectPublicKeyInfo', SubjectPublicKeyInfo),
        ('issuerUniqueID', Optional(Implicit(tag=1, typ=UniqueIdentifier))),
        ('subjectUniqueID', Optional(Implicit(tag=2, typ=UniqueIdentifier))),
        ('extensions', Optional(Explicit(tag=3, typ=Extensions))),
    ]


class Certificate(asn1.Seq):
    fields = [
        ('tbsCertificate', TBSCertificate),
        ('signatureAlgorithm', DigestAlgorithmIdentifier),
        ('signature', asn1.BitStr),
    ]

    @property
    def extensions(self):
        return self.tbsCertificate.extensions

    @property
    def subject_edrpou(self):
        tax = self.extensions['2.5.29.9']
        return str(tax['1.2.804.2.1.1.1.11.1.4.2.1'])

    @property
    def subject_drfo(self):
        tax = self.extensions['2.5.29.9']
        return str(tax['1.2.804.2.1.1.1.11.1.4.1.1'])



from __future__ import print_function
import datetime
import types
from . markers import Optional, Default, Implicit, Explicit
from . rewindable import Rewindable
from . util import u


class Impossible(object):
    pass

impossible = Impossible()

class Asn1Tag(object):
    primitive = True

    def __init__(self, data=None, decode_fn=None, value=None, tag=None):
        if data:
            assert not isinstance(data, Asn1Tag)
            self.read(data)
        else:
            self.value = value

    @classmethod
    def match(cls, tag):
        if cls.tag == tag:
            return cls

    @classmethod
    def from_data(cls, data, decode_fn=None, tag=None, **kwargs):
        if isinstance(data, cls) is cls:
            return data

        if issubclass(cls, type(data)):
            rt = cls(value=data.value)
            return rt

        return cls(data, decode_fn=decode_fn)

    @classmethod
    def to_data(cls, data, parent):
        return data

    @classmethod
    def stream(cls, data, tlen=None, decode_fn=None):
        if type(data) is not Rewindable:
            tlen = tlen or len(data)
            data = Rewindable(decode_fn(data, tlen))

        tag, clsname, content = data.next()
        if tag != cls.tag:
            raise ValueError("Schema mismcatch tag {}".format(tag))

        if type(content) is Rewindable:
            return cls(content, decode_fn)
        return content

    def encode_type(self, el, desc):
        tag = el.tag
        cls = 0
        content = el.encode()
        wrapTag = None
        if type(desc) is Explicit:
            wrapTag = desc.tag |  0x20
            desc = desc.typ

        if type(desc) is Implicit:
            tag = desc.tag
            cls = 2
            desc = desc.typ
            bconstructed = 0 if desc.primitive else 0x20
            tag |= bconstructed

        if type(desc) is types.FunctionType:
            desc = desc(self)

        ret = desc.to_data((tag, cls, content), self)
        if wrapTag is not None:
            ret = (wrapTag, 2, (ret,))

        return ret


class Seq(Asn1Tag):
    tag = 0x10
    primitive = False

    def __init__(self, source=None, decode_fn=None, tag=None, value=None):
        self.elements = []
        if type(source) is Rewindable:
            self.read(source, decode_fn)

        self.source = source

    def read(self, source, decode_fn):
        self.read_fields(source, decode_fn)

    def read_fields(self, source, decode_fn):
        fields = self.fields[:]

        while fields:
            is_optional = False
            is_default = False
            fallback = None
            name, desc = fields.pop(0)
            if type(desc) is types.FunctionType:
                desc = desc(self)

            if type(desc) is Optional:
                desc = desc.typ
                is_optional = True

            if type(desc) is Default:
                is_default = True
                fallback = desc.value
                desc = desc.typ

            try:
                tag, cls, content = source.next()
            except StopIteration:
                if is_default:
                    setattr(self, name, desc(value=fallback))
                    break

                if is_optional:
                    break

                raise ValueError("Incomplete structure")

            match_desc = desc.match(tag)
            if match_desc is None:
                if is_optional:
                    source.rewind((tag, cls, content))
                    continue
                elif is_default:
                    source.rewind((tag, cls, content))
                    setattr(self, name, desc(value=fallback))
                    continue
                else:
                    raise ValueError("Input doesnt match schema at '{}'. Want {}, see tag {}".format(name, desc, hex(tag)))
            desc = match_desc

            if type(desc) is Explicit:
                content = decode_fn(content.data, content.tlen, content.off)
                tag, cls, content = next(content)
                desc = desc.typ

            if type(desc) is Implicit:
                desc = desc.typ
                if issubclass(desc, (Seq, SetOf)):
                    content = decode_fn(content.data, content.tlen, content.off)
                else:
                    content = content.data[content.off:content.off+content.tlen]

            if type(desc) is types.FunctionType:
                desc = desc(self)

            content = desc.from_data(content, decode_fn, tag=tag, parent=self)
            setattr(self, name, content)
            self.elements.append(name)

    def to_stream(self, encode_fn):
        return encode_fn(((
            self.tag | 0x20, 0, self.gen_contents()
        ),))

    def encode(self, cls=0):
        return self.gen_contents()

    def gen_contents(self):
        fields = self.fields[:]

        while fields:
            name, desc = fields.pop(0)
            if type(desc) is types.FunctionType:
                desc = desc(self)

            is_optional = False
            is_default = False
            default_value = impossible

            if type(desc) is Optional:
                desc = desc.typ
                is_optional = True

            if type(desc) is Default:
                default_value = desc.value
                desc = desc.typ
                is_default = True

            try:
                el = getattr(self, name)
            except AttributeError:
                if is_optional or is_default:
                    continue
                raise

            if is_default and default_value == el.value:
                continue

            yield self.encode_type(el, desc)

    def __repr__(self):
        return '<Seq {} of {}>'.format(self.__class__.__name__, str.join(', ', self.elements))


class ObjId(Asn1Tag):
    tag = 0x06

    def read(self, data):
        current = data[0]
        numbers = [int(current / 40), current % 40]
        current = 0
        for n in data[1:]:
            current <<= 7
            current |= n & 0x7F

            if n & 0x80 == 0:
                numbers.append(current)
                current = 0

        self.value = numbers

    def __repr__(self):
        return '<ObjId {}>'.format(str(self))

    def __str__(self):
        return str.join('.', map(str,self.value))

    def encode(self):
        ret = bytearray()
        numbers = self.value[:]
        n1, n2 = numbers.pop(0), numbers.pop(0)
        ret.append((n1 * 40) + n2)
        while numbers:
            n = numbers.pop(0)
            if n == 0:
                ret.append(n)
                continue

            off = len(ret)
            while n:
                nb = n & 0x7F
                ret.insert(off, nb | 0x80)
                n >>= 7

            ret[-1] &= 0x7F

        return ret


class OctStr(Asn1Tag):
    tag = 0x04

    def read(self, data):
        self.value = u(data, 'latin1')

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, str.encode(str(self.value), 'hex'))

    def __str__(self):
        return str(self.value)

    def encode(self):
        return bytearray(self.value, 'latin1')


class Utf8Str(OctStr):
    tag = 0x0C
    def read(self, data):
        self.value = u(data, 'utf8')

    def encode(self):
        return bytearray(self.value, 'utf-8')

    def __repr__(self):
        return '<{} UTF8 {}>'.format(self.__class__.__name__, self.value.encode('utf8'))


class PrintStr(OctStr):
    tag = 0x13

class BitStr(OctStr):
    tag = 0x03


class Int(Asn1Tag):
    tag = 0x02

    def read(self, data):
        value = 0
        off = 0
        ln = len(data)
        while off < ln:
            value <<= 8
            value |= data[off]
            off += 1

        self.value = value

    def __repr__(self):
        return '<Int {}>'.format(self.value)

    def encode(self):
        if not self.value:
            return bytearray([0])

        ret = bytearray()
        value = self.value
        while value:
            ret.insert(0, value & 0xFF)
            value >>= 8
        return ret


class Bool(Asn1Tag):
    tag = 0x1

    def read(self, data):
        self.value = bool(data[0])

    def encode(self):
        ret = 0xFF if self.value else 00
        return bytearray([ret])


class SetOf(Asn1Tag):
    tag = 0x11
    typ = None
    primitive = False
    def __init__(self, data, decode_fn, tag=None, value=None):
        self.elements = []
        if data and self.typ:
            self.read(data, decode_fn)
        else:
            self.data = data

    def read(self, source, decode_fn):
        for tag, cls, content in source:
            content = self.typ.from_data(content, decode_fn, tag=tag, parent=self)
            self.elements.append(content)

    def __repr__(self):
        return '<SetOf {}: {}>'.format(
            self.typ, 
            str.join(', ', map(repr, self.elements))
        )

    def __getitem__(self, idx):
        return self.elements[idx]

    def to_stream(self, encode_fn):
        return encode_fn(((
            self.tag | 0x20, 0, self.gen_contents()
        ),))

    def encode(self, cls=0):
        return  self.gen_contents()

    def gen_contents(self):
        desc = self.typ
        for el in self.elements:
            yield self.encode_type(el, desc)


class SeqOf(SetOf, Seq):
    tag = 0x10


class Time(Asn1Tag):
    tag = 0x17

    def read(self, data):
        year = int(data[:2] or "0")
        mon = int(data[2:4] or "0")
        day = int(data[4:6] or "0")
        hour = int(data[6:8] or "0")
        nmin = int(data[8:10] or "0")
        nsec = int(data[10:12] or "0")
        
        if year < 70:
            year = 2000 + year
        else:
            year = 1900 + year

        self.value = datetime.datetime(year=year, month=mon, day=day,
                                       hour=hour, minute=nmin,
                                       second=nsec)

    def __repr__(self):
        return '<UTCTime {}>'.format(self.value)

    def encode(self):
        return bytearray(self.value.strftime('%y%m%d%H%M%SZ'), 'latin1')


class Any(object):
    primitive = True
    def __init__(self, data, decode_fn=None, tag=None, *args, **kwargs):
        if type(data) is Rewindable:
            data = data.data[data.off:data.tlen+data.off]
        self.data = data
        self.tag = tag
        self.primitive = (0x13 & tag) not in [0x10, 0x11]

    @classmethod
    def from_data(cls, data, *args, **kwargs):
        return cls(data, *args, **kwargs)

    @classmethod
    def to_data(cls, data, parent):
        return data

    def encode(self):
        if hasattr(self.data, 'encode'):
            return self.data.encode()
        return self.data

    @classmethod
    def match(cls, tag):
        return cls



class Tagged(object):
    def __init__(self, tag=None, data=None, off=0, tlen=0, typ=None):
        self.data = data
        self.off = off
        self.tlen = tlen
        self.tag = tag
        self.typ = typ

    def encode(self):
        return self.data[self.off:self.off + self.tlen]

    def match(self, tag):
        if self.tag == tag:
            return self


class Implicit(Tagged):
    pass

class Explicit(Tagged):
    pass


class Optional(object):
    def __init__(self, typ):
        self.typ = typ


class Default(object):
    def __init__(self, value, typ):
        self.value = value
        self.typ = typ


class Choice(object):
    @classmethod
    def match(cls, tag):
        for spec in cls.types:
            match = spec.match(tag)
            if match is not None:
                return match

    @classmethod
    def to_data(cls, triplet, parent):
        tag, _cls, content = triplet
        match = cls.match(tag)
        if match is None:
            raise ValueError("No matching spec for {} tag {}".format(content, tag))
        return match.to_data(triplet, parent)



class Rewindable(object):
    def __init__(self, source=None, data=None, fn=None, off=None, tlen=None):
        self.source = source or fn(data, tlen, off)
        self.back = None
        self.data = data
        self.off = off
        self.tlen = tlen

    def __iter__(self):
        while True:
            yield self.next()

    def next(self):
        if self.back:
            out = self.back
            self.back = None
            return out

        return next(self.source)

    def rewind(self, value):
        self.back = value




# coding=utf8
from __future__ import print_function
import os.path
from zozol import decode_transport, decode_ber
from zozol.schemas.pkcs7_dstszi import ContentInfo


def here(fname):
    dirname, _ = os.path.split(__file__)
    return os.path.join(dirname, fname)


def test_unwrap():
    data = open(here('signed1'), 'rb').read()
    rdata = open(here('signed1.r'), 'rb').read()
    document = decode_transport(data)
    assert document.body == rdata
    assert document.label == 'UA1_SIGN'


def test_headers():
    data = open(here('signed2'), 'rb').read()
    document = decode_transport(data)
    assert document.label == 'UA1_SIGN'
    assert document.headers['SUBJECT'] == u'Квитанція №1 (не прийнято)'
    assert document.body[0] == 0x30, repr(document.body[:100])


def test_broken_encoding():
    data = open(here('signed3'), 'rb').read()
    document = decode_transport(data)
    assert document.label == 'UA1_SIGN'


def test_enveloped():
    data = open(here('signed4'), 'rb').read()
    document = decode_transport(data)
    assert document.label == 'UA1_SIGN'

    msg = ContentInfo.stream(bytearray(document.body), len(data), decode_ber)
    document = decode_transport(bytearray(msg.content.contentInfo.content.value.encode('latin1')))
    assert document.label == 'UA1_CRYPT'
    assert document.headers.get('CERT_PEM') is not None

    try:
        msg = ContentInfo.stream(bytearray(document.body), len(data), decode_ber)
    except KeyError as e:
        assert str(e) == "'1.2.840.113549.1.7.3'"
    else:
        assert False, "Should raise key error"



# -*- coding: utf-8 -*-
from setuptools import setup

setup(
    name="zozol",
    version='0.2.1',
    url='https://github.com/muromec/zozol',
    description='Broken ASN1 for DSTU',
    author='Ilya Petrov',
    author_email='ilya.muromec@gmail.com',
    packages=["zozol", "zozol.schemas"],
    include_package_data=True,
    zip_safe=False,
    install_requires=[ ],
    tests_require=[
        'gost89',
        'nose',
    ],
    setup_requires=['nose>=1.0'],
    classifiers=[
        'Programming Language :: Python',
    ]
)



from .. import base as asn1
from .. markers import Implicit, Explicit, Optional, Choice
from . rfc3280_x509 import Certificate, DigestAlgorithmIdentifier, Name


class ContentInfo(asn1.Seq):
    def resolve_content(obj):
        return TYPES[str(obj.contentType)]

    fields = [
        ('contentType', asn1.ObjId),
        ('content', Optional(Explicit(tag=0, typ=resolve_content))),
    ]


class DigestAlgorithmIdentifiers(asn1.SetOf):
    typ = DigestAlgorithmIdentifier


class ExtendedCertificatesAndCertificates(asn1.SeqOf):
    typ = Certificate


class IssuerAndSerialNumber(asn1.Seq):
    fields = [
        ('issuer', Name),
        ('serialNumber', asn1.Int),
    ]



class SignerIdentifier(Choice):
    types = [
        IssuerAndSerialNumber,
        asn1.OctStr
    ]


class AttributeValues(asn1.SetOf):
    typ = asn1.Any


class Attribute(asn1.Seq):
    fields = [
        ('type', asn1.ObjId),
        ('values', AttributeValues),
    ]

class Attributes(asn1.SetOf):
    typ = Attribute
    HUMAN_NAME = {
        "contentType": "1.2.840.113549.1.9.3",
        "messageDigest": "1.2.840.113549.1.9.4",
        "signingTime ": "1.2.840.113549.1.9.5",
    }

    def _build_attr_cache(self):
        return {
            str(el.type): el.values
            for el in self.elements
        }


    def __getitem__(self, idx):
        try:
            kv = self._attr_cache
        except AttributeError:
            kv = self._build_attr_cache()
            self._attr_cache = kv

        return kv[idx]

    def __getattr__(self, idx):
        objid = self.HUMAN_NAME.get(idx)
        if objid is None:
            raise AttributeError("No attr {} defined".format(idx))

        try:
            return self[objid][0].data
        except KeyError:
            raise AttributeError("Attribute {} not present".format(objid))

class SignerInfo(asn1.Seq):
    fields = [
        ('version', asn1.Int),
        ('sid', SignerIdentifier),
        ('digestAlgorithm', DigestAlgorithmIdentifier),
        ('authenticatedAttributes', Optional(Implicit(tag=0, typ=Attributes))),
        ('digestEncryptionAlgorithm', DigestAlgorithmIdentifier),
        ('encryptedDigest', asn1.OctStr),
        ('unauthenticatedAttributes', Optional(Implicit(tag=1, typ=Attributes))),
    ]


class SignerInfos(asn1.SetOf):
    typ = SignerInfo


class SignedData(asn1.Seq):
    fields = [
        ('version', asn1.Int),
        ('digestAlgorithms', DigestAlgorithmIdentifiers),
        ('contentInfo', ContentInfo),
        ('certificates', Optional(Implicit(tag=0, typ=ExtendedCertificatesAndCertificates))),
        ('crls', Optional(Implicit(tag=1, typ=asn1.Any))),
        ('signerInfos', SignerInfos),

    ]

class Data(asn1.OctStr):
    pass

TYPES = {
    "1.2.840.113549.1.7.2": SignedData,
    "1.2.840.113549.1.7.1": Data,
}

